# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import bisect
import itertools
from typing import Any, Callable, Generic, Iterator, List, Optional, Tuple, TypeVar

SortedListItem = TypeVar("SortedListItem")
SortedListKey = TypeVar("SortedListKey")


class SortedList(Generic[SortedListKey, SortedListItem]):
    data: List[Tuple[SortedListKey, SortedListItem]]

    # https://github.com/python/mypy/issues/708
    # key: Callable[[SortedListItem], SortedListKey]

    def __init__(
        self,
        data: List[SortedListItem] = [],
        key: Optional[Callable[[SortedListItem], SortedListKey]] = None,
    ):
        if key is None:

            def key(item):
                return item

        assert key is not None  # for mypy
        self.data = sorted((key(x), x) for x in data or [])

        self.key: Callable[[SortedListItem], SortedListKey] = key

    def add(self, item: SortedListItem):
        k = self.key(item)
        bisect.insort(self.data, (k, item))

    def __iter__(self) -> Iterator[SortedListItem]:
        for k, item in self.data:
            yield item

    def iter_from(self, start_key: Any) -> Iterator[SortedListItem]:
        """Returns an iterator over all the elements whose key is greater
        or equal to `start_key`.
        (This is an efficient equivalent to:
        `(x for x in L if key(x) >= start_key)`)
        """
        from_index = bisect.bisect_left(self.data, (start_key,))
        for k, item in itertools.islice(self.data, from_index, None):
            yield item

    def iter_after(self, start_key: Any) -> Iterator[SortedListItem]:
        """Same as iter_from, but using a strict inequality."""
        it = self.iter_from(start_key)
        for item in it:
            if self.key(item) > start_key:
                yield item
                break

        yield from it



# Copyright (C) 2019-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from importlib.metadata import distribution
import logging
import os
import sys
from typing import Dict, List, Optional

logger = logging.getLogger(__name__)


def get_sentry_release(
    main_package: Optional[str] = None, sentry_dsn: Optional[str] = None
):
    main_package = os.environ.get("SWH_MAIN_PACKAGE", main_package)
    if main_package:
        version = distribution(main_package).version
        return f"{main_package}@{version}"
    elif sentry_dsn is None:
        # return a dummy release when sentry_dsn is None to avoid side effects
        # related to sentry_sdk calling git command for fetching release info
        return "0.0.0"
    else:
        return None


def override_with_bool_envvar(envvar: str, default: bool) -> bool:
    """Override the `default` with the environment variable `envvar` parsed as a boolean"""
    envvalue = os.environ.get(envvar, "")
    if envvalue.lower() in ("t", "true", "y", "yes", "1"):
        return True
    elif envvalue.lower() in ("f", "false", "n", "no", "0"):
        return False
    else:
        if envvalue:
            logger.warning(
                "Could not interpret environment variable %s=%r as boolean, "
                "using default value %s",
                envvar,
                envvalue,
                default,
            )
        return default


def override_with_float_envvar(
    envvar: str, default: Optional[float]
) -> Optional[float]:
    """Override `default` with the environment variable `envvar` casted as a float.

    `default` is returned if the environment variable `envvar` is missing or if
    we're not able to cast it to a float.

    Args:
        envvar: the name of the environment variable
        default: default value

    Returns:
        A float or `default`
    """
    envvalue = os.environ.get(envvar)
    if envvalue is None:
        return default
    try:
        return float(envvalue)
    except ValueError:
        logger.warning(
            "Could not interpret environment variable %s=%r as float, "
            "using default value %s",
            envvar,
            envvalue,
            default,
        )
        return default


def init_sentry(
    sentry_dsn: Optional[str] = None,
    *,
    main_package: Optional[str] = None,
    environment: Optional[str] = None,
    debug: bool = False,
    disable_logging_events: bool = False,
    integrations: Optional[List] = None,
    traces_sample_rate: Optional[float] = None,
    extra_kwargs: Optional[Dict] = None,
    deferred_init: bool = False,
) -> None:
    """Configure the sentry integration.

    Args:
      sentry_dsn: Sentry DSN; where sentry report will be sent. Overridden by
        :envvar:`SWH_SENTRY_DSN`
      main_package: Full name of main Python package associated to Sentry DSN.
        Overridden by :envvar:`SWH_MAIN_PACKAGE`.
      environment: Sentry environment. Overridden by :envvar:`SWH_SENTRY_ENVIRONMENT`
      debug: turn on Sentry SDK debug mode. Overridden by :envvar:`SWH_SENTRY_DEBUG`
      disable_logging_events: if set, disable the automatic reporting of error/exception
        log entries as Sentry events. Overridden by
        :envvar:`SWH_SENTRY_DISABLE_LOGGING_EVENTS`
      integrations: list of dedicated Sentry integrations to include
      traces_sample_rate: a number between 0 and 1, controlling the percentage chance a
        given transaction will be sent to Sentry. Overridden by
        :envvar:`SWH_SENTRY_TRACES_SAMPLE_RATE`
      extra_kwargs: dict of additional parameters passed to :func:`sentry_sdk.init`
      deferred_init: indicates that sentry will be properly initialized in subsequent
        calls and that no warnings about missing DSN should be logged
    """
    if integrations is None:
        integrations = []
    if extra_kwargs is None:
        extra_kwargs = {}

    sentry_dsn = os.environ.get("SWH_SENTRY_DSN", sentry_dsn)
    environment = os.environ.get("SWH_SENTRY_ENVIRONMENT", environment)

    debug = override_with_bool_envvar("SWH_SENTRY_DEBUG", debug)
    disable_logging_events = override_with_bool_envvar(
        "SWH_SENTRY_DISABLE_LOGGING_EVENTS", disable_logging_events
    )

    if sentry_dsn is None and not deferred_init:
        # Don’t display a warning if there is a controlling terminal
        # as errors can be monitored from there.
        if not sys.stdout.isatty():
            logger.warning("Sentry DSN not provided, events will not be sent.")

    import sentry_sdk

    if disable_logging_events:
        from sentry_sdk.integrations.logging import LoggingIntegration

        integrations.append(LoggingIntegration(event_level=None))

    # to completely disable tracing `traces_sample_rate` should be set to None instead
    # of 0.0
    traces_sample_rate = override_with_float_envvar(
        "SWH_SENTRY_TRACES_SAMPLE_RATE", traces_sample_rate
    )

    sentry_sdk.init(
        release=get_sentry_release(main_package, sentry_dsn),
        environment=environment,
        dsn=sentry_dsn,
        traces_sample_rate=traces_sample_rate,
        integrations=integrations,
        debug=debug,
        **extra_kwargs,
    )



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Initially imported from https://github.com/DataDog/datadogpy/
# at revision 62b3a3e89988dc18d78c282fe3ff5d1813917436
#
# Copyright (c) 2015, Datadog <info@datadoghq.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Datadog nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
#
# Vastly adapted for integration in swh.core:
#
# - Removed python < 3.5 compat code
# - trimmed the imports down to be a single module
# - adjust some options:
#   - drop unix socket connection option
#   - add environment variable support for setting the statsd host and
#     port (pulled the idea from the main python statsd module)
#   - only send timer metrics in milliseconds (that's what
#     prometheus-statsd-exporter expects)
#   - drop DataDog-specific metric types (that are unsupported in
#     prometheus-statsd-exporter)
# - made the tags a dict instead of a list (prometheus-statsd-exporter only
#   supports tags with a value, mirroring prometheus)
# - switch from time.time to time.monotonic
# - improve unit test coverage
# - documentation cleanup


from asyncio import iscoroutinefunction
from contextlib import contextmanager
from functools import wraps
import itertools
import logging
import os
from random import random
import re
import socket
import threading
from time import monotonic
from typing import Collection, Dict, Optional
import warnings

log = logging.getLogger("swh.core.statsd")


class TimedContextManagerDecorator(object):
    """
    A context manager and a decorator which will report the elapsed time in
    the context OR in a function call.

    Attributes:
      elapsed (float): the elapsed time at the point of completion
    """

    def __init__(
        self, statsd, metric=None, error_metric=None, tags=None, sample_rate=1
    ):
        self.statsd = statsd
        self.metric = metric
        self.error_metric = error_metric
        self.tags = tags or {}
        self.sample_rate = sample_rate
        self.elapsed = None  # this is for testing purpose

    def __call__(self, func):
        """
        Decorator which returns the elapsed time of the function call.

        Default to the function name if metric was not provided.
        """
        if not self.metric:
            self.metric = "%s.%s" % (func.__module__, func.__name__)

        # Coroutines
        if iscoroutinefunction(func):

            @wraps(func)
            async def wrapped_co(*args, **kwargs):
                start = monotonic()
                try:
                    result = await func(*args, **kwargs)
                except BaseException as e:
                    self._send_error(error_type=type(e).__name__)
                    raise
                self._send(start)
                return result

            return wrapped_co

        # Others
        @wraps(func)
        def wrapped(*args, **kwargs):
            start = monotonic()
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                self._send_error(error_type=type(e).__name__)
                raise
            self._send(start)
            return result

        return wrapped

    def __enter__(self):
        if not self.metric:
            raise TypeError("Cannot used timed without a metric!")
        self._start = monotonic()
        return self

    def __exit__(self, type, value, traceback):
        # Report the elapsed time of the context manager if no error.
        if type is None:
            self._send(self._start)
        else:
            self._send_error(error_type=type.__name__)

    def _send(self, start):
        elapsed = (monotonic() - start) * 1000
        self.statsd.timing(
            self.metric, elapsed, tags=self.tags, sample_rate=self.sample_rate
        )
        self.elapsed = elapsed

    def _send_error(self, error_type=None):
        if self.error_metric is None:
            self.error_metric = self.metric + "_error_count"
        if error_type is not None:
            tags = {**self.tags, "error_type": error_type}
        else:
            tags = self.tags
        self.statsd.increment(self.error_metric, tags=tags)

    def start(self):
        """Start the timer"""
        self.__enter__()

    def stop(self):
        """Stop the timer, send the metric value"""
        self.__exit__(None, None, None)


class Statsd(object):
    """Initialize a client to send metrics to a StatsD server.

    Arguments:
      host (str): the host of the StatsD server. Defaults to localhost.
      port (int): the port of the StatsD server. Defaults to 8125.

      max_buffer_size (int): Maximum number of metrics to buffer before
        sending to the server if sending metrics in batch

      namespace (str): Namespace to prefix all metric names

      constant_tags (Dict[str, str]): Tags to attach to all metrics

    Note:
      This class also supports the following environment variables:

      STATSD_HOST
        Override the default host of the statsd server
      STATSD_PORT
        Override the default port of the statsd server
      STATSD_TAGS
        Tags to attach to every metric reported. Example value:

        "label:value,other_label:other_value"
    """

    def __init__(
        self,
        *,
        host=None,
        port=None,
        max_buffer_size=50,
        namespace=None,
        constant_tags=None,
    ):
        # Connection
        if host is None:
            host = os.environ.get("STATSD_HOST") or "localhost"
        self.host = host

        if port is None:
            port = os.environ.get("STATSD_PORT") or 8125
        self.port = int(port)

        # Socket
        self._socket = None
        self.lock = threading.Lock()
        self.max_buffer_size = max_buffer_size
        self._send = self._send_to_server
        self.encoding = "utf-8"

        # Tags
        self.constant_tags = {}
        tags_envvar = os.environ.get("STATSD_TAGS", "")
        for tag in tags_envvar.split(","):
            if not tag:
                continue
            if ":" not in tag:
                warnings.warn(
                    f"STATSD_TAGS needs to be in 'key:value' format, not {tag!r}",
                    UserWarning,
                )
                continue
            k, v = tag.split(":", 1)

            # look for a possible env var substitution, using $NAME or ${NAME} format
            m = re.match(r"^[$]([{])?(?P<envvar>\w+)(?(1)[}]|)$", v)
            if m:
                envvar = m.group("envvar")
                if envvar in os.environ:
                    v = os.environ[envvar]

            self.constant_tags[k] = v

        if constant_tags:
            self.constant_tags.update(
                {str(k): str(v) for k, v in constant_tags.items()}
            )

        # Namespace
        if namespace is not None:
            namespace = str(namespace)
        self.namespace = namespace

    def __enter__(self):
        self.open_buffer(self.max_buffer_size)
        return self

    def __exit__(self, type, value, traceback):
        self.close_buffer()

    def gauge(self, metric, value, tags=None, sample_rate=1):
        """
        Record the value of a gauge, optionally setting a list of tags and a
        sample rate.

        >>> statsd.gauge('users.online', 123)
        >>> statsd.gauge('active.connections', 1001, tags={"protocol": "http"})
        """
        return self._report(metric, "g", value, tags, sample_rate)

    def increment(self, metric, value=1, tags=None, sample_rate=1):
        """
        Increment a counter, optionally setting a value, tags and a sample
        rate.

        >>> statsd.increment('page.views')
        >>> statsd.increment('files.transferred', 124)
        """
        self._report(metric, "c", value, tags, sample_rate)

    def decrement(self, metric, value=1, tags=None, sample_rate=1):
        """
        Decrement a counter, optionally setting a value, tags and a sample
        rate.

        >>> statsd.decrement('files.remaining')
        >>> statsd.decrement('active.connections', 2)
        """
        metric_value = -value if value else value
        self._report(metric, "c", metric_value, tags, sample_rate)

    def histogram(self, metric, value, tags=None, sample_rate=1):
        """
        Sample a histogram value, optionally setting tags and a sample rate.

        >>> statsd.histogram('uploaded.file.size', 1445)
        >>> statsd.histogram('file.count', 26, tags={"filetype": "python"})
        """
        self._report(metric, "h", value, tags, sample_rate)

    def timing(self, metric, value, tags=None, sample_rate=1):
        """
        Record a timing, optionally setting tags and a sample rate.

        >>> statsd.timing("query.response.time", 1234)
        """
        self._report(metric, "ms", value, tags, sample_rate)

    def timed(self, metric=None, error_metric=None, tags=None, sample_rate=1):
        """
        A decorator or context manager that will measure the distribution of a
        function's/context's run time. Optionally specify a list of tags or a
        sample rate. If the metric is not defined as a decorator, the module
        name and function name will be used. The metric is required as a
        context manager.
        ::

            @statsd.timed('user.query.time', sample_rate=0.5)
            def get_user(user_id):
                # Do what you need to ...
                pass

            # Is equivalent to ...
            with statsd.timed('user.query.time', sample_rate=0.5):
                # Do what you need to ...
                pass

            # Is equivalent to ...
            start = time.monotonic()
            try:
                get_user(user_id)
            finally:
                statsd.timing('user.query.time', time.monotonic() - start)
        """
        return TimedContextManagerDecorator(
            statsd=self,
            metric=metric,
            error_metric=error_metric,
            tags=tags,
            sample_rate=sample_rate,
        )

    def set(self, metric, value, tags=None, sample_rate=1):
        """
        Sample a set value.

        >>> statsd.set('visitors.uniques', 999)
        """
        self._report(metric, "s", value, tags, sample_rate)

    @property
    def socket(self):
        """
        Return a connected socket.

        Note: connect the socket before assigning it to the class instance to
        avoid bad thread race conditions.
        """
        with self.lock:
            if not self._socket:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.connect((self.host, self.port))
                self._socket = sock

            return self._socket

    def open_buffer(self, max_buffer_size=50):
        """
        Open a buffer to send a batch of metrics in one packet.

        You can also use this as a context manager.

        >>> with Statsd() as batch:
        ...     batch.gauge('users.online', 123)
        ...     batch.gauge('active.connections', 1001)
        """
        self.max_buffer_size = max_buffer_size
        self.buffer = []
        self._send = self._send_to_buffer

    def close_buffer(self):
        """
        Flush the buffer and switch back to single metric packets.
        """
        self._send = self._send_to_server

        if self.buffer:
            # Only send packets if there are packets to send
            self._flush_buffer()

    def close_socket(self):
        """
        Closes connected socket if connected.
        """
        with self.lock:
            if self._socket:
                self._socket.close()
                self._socket = None

    def _report(self, metric, metric_type, value, tags, sample_rate):
        """
        Create a metric packet and send it.
        """
        if value is None:
            return

        if sample_rate != 1 and random() > sample_rate:
            return

        # Resolve the full tag list
        tags = self._add_constant_tags(tags)

        # Create/format the metric packet
        payload = "%s%s:%s|%s%s%s" % (
            (self.namespace + ".") if self.namespace else "",
            metric,
            value,
            metric_type,
            ("|@" + str(sample_rate)) if sample_rate != 1 else "",
            (
                ("|#" + ",".join("%s:%s" % (k, v) for (k, v) in sorted(tags.items())))
                if tags
                else ""
            ),
        )
        # Send it
        self._send(payload)

    def _send_to_server(self, packet):
        try:
            # If set, use socket directly
            self.socket.send(packet.encode("utf-8"))
        except socket.timeout:
            return
        except socket.error:
            log.debug(
                "Error submitting statsd packet."
                " Dropping the packet and closing the socket."
            )
            self.close_socket()

    def _send_to_buffer(self, packet):
        self.buffer.append(packet)
        if len(self.buffer) >= self.max_buffer_size:
            self._flush_buffer()

    def _flush_buffer(self):
        self._send_to_server("\n".join(self.buffer))
        self.buffer = []

    def _add_constant_tags(self, tags):
        return {
            str(k): str(v)
            for k, v in itertools.chain(
                self.constant_tags.items(),
                (tags if tags else {}).items(),
            )
        }

    @contextmanager
    def status_gauge(
        self,
        metric_name: str,
        statuses: Collection[str],
        tags: Optional[Dict[str, str]] = None,
    ):
        """Context manager to keep track of status changes as a gauge

        In addition to the `metric_name` and `tags` arguments, it expects a
        list of `statuses` to declare which statuses are possible, and returns
        a callable as context manager. This callable takes ones of the possible
        statuses as argument. Typical usage would be:

        >>> with statsd.status_gauge(
                    "metric_name", ["starting", "processing", "waiting"]) as set_status:
                set_status("starting")
                # ...
                set_status("waiting")
                # ...
        """
        if tags is None:
            tags = {}
        current_status: Optional[str] = None
        # reset status gauges to make sure they do not "leak"
        for status in statuses:
            self.gauge(metric_name, 0, {**tags, "status": status})

        def set_status(new_status: str):
            nonlocal current_status
            assert isinstance(tags, dict)
            if new_status not in statuses:
                raise ValueError(f"{new_status} not in {statuses}")
            if current_status and new_status != current_status:
                self.gauge(metric_name, 0, {**tags, "status": current_status})
            current_status = new_status
            self.gauge(metric_name, 1, {**tags, "status": current_status})

        try:
            yield set_status
        finally:
            # reset gauges on exit
            for status in statuses:
                self.gauge(metric_name, 0, {**tags, "status": status})


statsd = Statsd()



# Copyright (C) 2023-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import pytest
import requests
from requests.status_codes import codes
from tenacity.wait import wait_fixed

from swh.core.retry import MAX_NUMBER_ATTEMPTS, WAIT_EXP_BASE, http_retry

TEST_URL = "https://example.og/api/repositories"


@http_retry()
def make_request():
    response = requests.get(TEST_URL)
    response.raise_for_status()
    return response


@pytest.fixture
def mock_sleep(mocker):
    return mocker.patch("time.sleep")


def assert_sleep_calls(mocker, mock_sleep, sleep_params):
    mock_sleep.assert_has_calls([mocker.call(param) for param in sleep_params])


@pytest.mark.parametrize(
    "status_code",
    [
        codes.too_many_requests,
        codes.internal_server_error,
        codes.bad_gateway,
        codes.service_unavailable,
    ],
)
def test_http_retry(requests_mock, mocker, mock_sleep, status_code):
    data = {"result": {}}
    requests_mock.get(
        TEST_URL,
        [
            {"status_code": status_code},
            {"status_code": status_code},
            {"status_code": codes.ok, "json": data},
        ],
    )

    response = make_request()

    assert_sleep_calls(mocker, mock_sleep, [1, WAIT_EXP_BASE])

    assert response.json() == data


def test_http_retry_max_attemps(requests_mock, mocker, mock_sleep):
    requests_mock.get(
        TEST_URL,
        [{"status_code": codes.too_many_requests}] * (MAX_NUMBER_ATTEMPTS),
    )

    with pytest.raises(requests.exceptions.HTTPError) as e:
        make_request()

    assert e.value.response.status_code == codes.too_many_requests

    assert_sleep_calls(
        mocker,
        mock_sleep,
        [float(WAIT_EXP_BASE**i) for i in range(MAX_NUMBER_ATTEMPTS - 1)],
    )


@http_retry(wait=wait_fixed(WAIT_EXP_BASE))
def make_request_wait_fixed():
    response = requests.get(TEST_URL)
    response.raise_for_status()
    return response


def test_http_retry_wait_fixed(requests_mock, mocker, mock_sleep):
    requests_mock.get(
        TEST_URL,
        [
            {"status_code": codes.too_many_requests},
            {"status_code": codes.too_many_requests},
            {"status_code": codes.ok},
        ],
    )

    make_request_wait_fixed()

    assert_sleep_calls(mocker, mock_sleep, [WAIT_EXP_BASE] * 2)



#!/usr/bin/env python3
# Copyright (C) 2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from os import get_terminal_size

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

logger = logging.getLogger(__name__)


@swh_cli_group.group(name="backend", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def backend(ctx):
    """Software Heritage backend generic tools."""
    pass


@backend.command(name="list", context_settings=CONTEXT_SETTINGS)
@click.argument("package", required=True)
@click.argument("cls", required=False, default=None)
@click.pass_context
def bk_list(ctx, package, cls):
    """Show registered backends for the given package

    With their documentation, if any. Example::

        \b
        $ swh backend list vault

        memory     Stub vault backend, for use in the CLI.
        postgresql Backend for the Software Heritage Vault.
        remote     Client to the Software Heritage vault cache.

    If 'cls' is given, display the full docstring for the corresponding
    backend.

    Example::

        \b
        $ swh backend list vault memory

        vault:memory

        Stub vault backend, for use in the CLI.

    """
    from swh.core.config import get_swh_backend_module, list_swh_backends

    if cls is None:
        items = []
        for backend in list_swh_backends(package):
            _, BackendCls = get_swh_backend_module(package, backend)
            msg = BackendCls.__doc__
            if msg is None:
                msg = ""
            msg = msg.strip()
            if "\n" in msg:
                firstline = msg.splitlines()[0]
            else:
                firstline = msg
            items.append((backend, firstline, msg))
        if not items:
            click.secho(
                f"No backend found for package '{package}'",
                fg="red",
                bold=True,
                err=True,
            )
            raise click.Abort()

        max_name = max(len(name) for name, _, _ in items)
        try:
            width = get_terminal_size().columns
        except OSError:
            width = 78
        for name, firstline, msg in items:
            click.echo(
                click.style(
                    f"{name:<{max_name + 1}}",
                    fg="green",
                    bold=True,
                ),
                nl=False,
            )
            firstline = firstline[: width - max_name - 1]
            click.echo(firstline)
    else:
        try:
            _, BackendCls = get_swh_backend_module(package, cls)
        except ValueError:
            BackendCls = None

        if BackendCls is None:
            click.secho(
                f"No backend '{cls}' found for package '{package}'",
                fg="red",
                bold=True,
                err=True,
            )
            raise click.Abort()

        click.echo(
            click.style(
                package,
                fg="green",
                bold=True,
            )
            + ":"
            + click.style(
                cls,
                fg="yellow",
                bold=True,
            )
            + "\n",
        )
        click.echo(BackendCls.__doc__.strip())



# Copyright (C) 2020-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


import logging
import random
import re
import time
from typing import Any, Dict, List, Optional

import requests
from requests.exceptions import ChunkedEncodingError, ConnectionError
from tenacity import (
    retry,
    retry_any,
    retry_if_exception_type,
    retry_if_result,
    stop_after_attempt,
    wait_exponential,
)

from ..statsd import Statsd

GITHUB_PATTERN = re.compile(
    r"(//|git://|git@|git//|https?://|ssh://|.*@)github.com[/:](?P<user_repo>.*)"
)


logger = logging.getLogger(__name__)


def _url_github_api(user_repo: str) -> str:
    """Given the user_repo, returns the expected github api url."""
    return f"https://api.github.com/repos/{user_repo}"


_SANITIZATION_RE = re.compile(r"^(.*?)/?(\.git)?/?$")


def _sanitize_github_url(url: str) -> str:
    """Sanitize github url."""
    m = _SANITIZATION_RE.match(url.lower())
    assert m is not None, url  # impossible, but mypy doesn't know it
    return m.group(1)


def get_canonical_github_origin_url(
    url: str, credentials: Optional[List[Dict[str, str]]] = None
) -> Optional[str]:
    """Retrieve canonical github url out of an url if any or None otherwise.

    This triggers an http request to the github api url to determine the canonical
    repository url (if no credentials is provided, the http request is anonymous. Either
    way that request can be rate-limited by github.)

    """
    return GitHubSession(
        user_agent="SWH core library", credentials=credentials
    ).get_canonical_url(url)


class RateLimited(Exception):
    def __init__(self, response):
        self.reset_time: Optional[int]

        # Figure out how long we need to sleep because of that rate limit
        ratelimit_reset = response.headers.get("X-Ratelimit-Reset")
        retry_after = response.headers.get("Retry-After")
        if ratelimit_reset is not None:
            self.reset_time = int(ratelimit_reset)
        elif retry_after is not None:
            self.reset_time = int(time.time()) + int(retry_after) + 1
        else:
            logger.warning(
                "Received a rate-limit-like status code %s, but no rate-limit "
                "headers set. Response content: %s",
                response.status_code,
                response.content,
            )
            self.reset_time = None
        self.response = response


class MissingRateLimitReset(Exception):
    pass


class GitHubSession:
    """Manages a :class:`requests.Session` with (optionally) multiple credentials,
    and cycles through them when reaching rate-limits."""

    credentials: Optional[List[Dict[str, str]]] = None

    def __init__(
        self, user_agent: str, credentials: Optional[List[Dict[str, str]]] = None
    ) -> None:
        """Initialize a requests session with the proper headers for requests to
        GitHub."""
        if credentials:
            creds = credentials.copy()
            random.shuffle(creds)
            self.credentials = creds

        self.statsd = Statsd(
            namespace="swh_outbound_api",
            constant_tags={"api_type": "github", "api_instance": "github"},
        )

        self.session = requests.Session()

        self.session.headers.update(
            {"Accept": "application/vnd.github.v3+json", "User-Agent": user_agent}
        )

        self.anonymous = not self.credentials

        if self.anonymous:
            logger.warning("No tokens set in configuration, using anonymous mode")

        self.token_index = -1
        self.current_user: Optional[str] = None

        if not self.anonymous:
            # Initialize the first token value in the session headers
            self.set_next_session_token()

    def set_next_session_token(self) -> None:
        """Update the current authentication token with the next one in line."""

        assert self.credentials

        self.token_index = (self.token_index + 1) % len(self.credentials)

        auth = self.credentials[self.token_index]

        self.current_user = auth["username"]
        logger.debug("Using authentication token for user %s", self.current_user)

        if "password" in auth:
            token = auth["password"]
        else:
            token = auth["token"]

        self.session.headers.update({"Authorization": f"token {token}"})

    @retry(
        wait=wait_exponential(multiplier=1, min=4, max=10),
        stop=stop_after_attempt(5),
        retry=retry_any(
            # ChunkedEncodingError happens when the TLS connection gets reset, e.g. when
            # running the lister on a connection with high latency.
            retry_if_exception_type(ChunkedEncodingError),
            # ConnectionError happen when the server hangs up
            retry_if_exception_type(ConnectionError),
            # 502 status codes happen for a Server Error, sometimes
            retry_if_result(lambda r: r.status_code == 502),
        ),
    )
    def _request(self, url: str) -> requests.Response:
        # When anonymous, rate-limits are per-IP; but we cannot necessarily
        # get the IP/hostname here as we may be containerized. Instead, we rely on
        # statsd-exporter adding the hostname in the 'instance' tag.
        tags = {"username": self.current_user or "anonymous"}

        self.statsd.increment("requests_total", tags=tags)
        response = self.session.get(url)

        # self.session.get(url) raises in case of non-HTTP error (DNS, TCP, TLS, ...),
        # so responses_total may differ from requests_total.
        self.statsd.increment(
            "responses_total", tags={**tags, "http_status": response.status_code}
        )

        try:
            ratelimit_remaining = int(response.headers["x-ratelimit-remaining"])
        except (KeyError, ValueError):
            logger.warning(
                "Invalid x-ratelimit-remaining header from GitHub: %r",
                response.headers.get("x-ratelimit-remaining"),
            )
        else:
            self.statsd.gauge("remaining_requests", ratelimit_remaining, tags=tags)

        try:
            reset_seconds = int(response.headers["x-ratelimit-reset"]) - time.time()
        except (KeyError, ValueError):
            logger.warning(
                "Invalid x-ratelimit-reset header from GitHub: %r",
                response.headers.get("x-ratelimit-reset"),
            )
        else:
            self.statsd.gauge("reset_seconds", reset_seconds, tags=tags)

        if (
            # 429 status code was previously returned for authenticated users, keep the
            # check for backward compatibility
            response.status_code == 429
            or (
                # https://docs.github.com/en/rest/overview
                # /resources-in-the-rest-api?apiVersion=2022-11-28#exceeding-the-rate-limit
                response.status_code == 403
                and response.json()
                .get("message", "")
                .startswith("API rate limit exceeded")
            )
        ):
            self.statsd.increment("rate_limited_responses_total", tags=tags)
            raise RateLimited(response)

        return response

    def request(self, url) -> requests.Response:
        """Repeatedly requests the given URL, cycling through credentials and sleeping
        if necessary; until either a successful response or :exc:`MissingRateLimitReset`
        """
        # The following for/else loop handles rate limiting; if successful,
        # it provides the rest of the function with a `response` object.
        #
        # If all tokens are rate-limited, we sleep until the reset time,
        # then `continue` into another iteration of the outer while loop,
        # attempting to get data from the same URL again.

        while True:
            max_attempts = len(self.credentials) if self.credentials else 1
            reset_times: Dict[int, int] = {}  # token index -> time
            for attempt in range(max_attempts):
                try:
                    return self._request(url)
                except RateLimited as e:
                    reset_info = "(unknown reset)"
                    if e.reset_time is not None:
                        reset_times[self.token_index] = e.reset_time
                        reset_info = "(resetting in %ss)" % (e.reset_time - time.time())

                    if not self.anonymous:
                        logger.info(
                            "Rate limit exhausted for current user %s %s",
                            self.current_user,
                            reset_info,
                        )
                        # Use next token in line
                        self.set_next_session_token()
                        # Wait one second to avoid triggering GitHub's abuse rate limits
                        self.statsd.increment("sleep_seconds_total", 1)
                        time.sleep(1)

            # All tokens have been rate-limited. What do we do?

            if not reset_times:
                logger.warning(
                    "No X-Ratelimit-Reset value found in responses for any token; "
                    "Giving up."
                )
                raise MissingRateLimitReset()

            sleep_time = max(reset_times.values()) - time.time() + 1
            logger.info(
                "Rate limits exhausted for all tokens. Sleeping for %f seconds.",
                sleep_time,
            )
            self.statsd.increment("sleep_seconds_total", sleep_time)
            time.sleep(sleep_time)

    def get_repository_metadata(self, repo_url: str) -> Optional[Dict[str, Any]]:
        """Retrieve metadata of a repository from the github API.

        Args:
            repo_url: URL of a github repository

        Returns:
            A dictionary holding the metadata of the repository or None
            if this is not a valid github repository.

        Throws:
            requests.HTTPError: if the request to the github API failed.
        """
        url = repo_url.lower()

        match = GITHUB_PATTERN.match(url)
        if not match:
            return None

        user_repo = _sanitize_github_url(match.groupdict()["user_repo"])
        response = self.request(_url_github_api(user_repo))
        response.raise_for_status()
        return response.json()

    def get_canonical_url(self, repo_url: str) -> Optional[str]:
        """Retrieve canonical github url out of a github url.

        This triggers an http request to the github api url to determine the
        canonical repository url.

        Args:
            repo_url: URL of a github repository

        Returns:
            The canonical github url, the input url if it is not a github one,
            None otherwise.
        """

        try:
            metadata = self.get_repository_metadata(repo_url)
            return metadata.get("html_url") if metadata else repo_url
        except requests.HTTPError:
            # invalid github repository
            return None



# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import itertools
import logging
from unittest.mock import call, patch

import pytest
import requests
from requests.exceptions import ChunkedEncodingError, ConnectionError
from tenacity import RetryError

from swh.core.github.pytest_plugin import HTTP_GITHUB_API_URL
from swh.core.github.utils import (
    GitHubSession,
    _sanitize_github_url,
    _url_github_api,
    get_canonical_github_origin_url,
)

KNOWN_GH_REPO = "https://github.com/user/repo"
KNOWN_GH_REPO2 = "https://github.com/user/reposit"


@patch("time.sleep", return_value=None)
@patch("swh.core.github.utils.requests.Session.get")
@pytest.mark.parametrize("exception", [ChunkedEncodingError, ConnectionError])
def test_retry(mock_requests, monkeypatch_sleep_calls, exception):
    user_repo = "test/test"
    html_input_url = f"https://github.com/{user_repo}"

    mock_requests.side_effect = exception("Request failure")

    with pytest.raises(RetryError, match=exception.__name__):
        get_canonical_github_origin_url(html_input_url)


@pytest.mark.parametrize(
    "user_repo, expected_url",
    [
        ("user/repo.git", KNOWN_GH_REPO),
        ("user/repo.git/", KNOWN_GH_REPO),
        ("user/repo/", KNOWN_GH_REPO),
        ("user/repo", KNOWN_GH_REPO),
        ("user/repo/.git", KNOWN_GH_REPO),
        ("user/reposit.git", KNOWN_GH_REPO2),
        ("user/reposit.git/", KNOWN_GH_REPO2),
        ("user/reposit/", KNOWN_GH_REPO2),
        ("user/reposit", KNOWN_GH_REPO2),
        ("user/reposit/.git", KNOWN_GH_REPO2),
        ("unknown/page", None),  # unknown gh origin returns None
        ("user/with/deps", None),  # url kind is not dealt with
    ],
)
def test_get_canonical_github_origin_url(
    user_repo, expected_url, requests_mock, github_credentials
):
    """It should return a canonical github origin when it exists, None otherwise"""
    for separator in ["/", ":"]:
        for prefix in [
            "http://",
            "https://",
            "git://",
            "ssh://",
            "//",
            "git@",
            "ssh://git@",
            "https://${env.GITHUB_TOKEN_USR}:${env.GITHUB_TOKEN_PSW}@",
            "[fetch=]git@",
        ]:
            html_input_url = f"{prefix}github.com{separator}{user_repo}"
            html_url = f"https://github.com/{user_repo}"
            api_url = _url_github_api(_sanitize_github_url(user_repo))

            if expected_url is not None:
                status_code = 200
                response = {"html_url": _sanitize_github_url(html_url)}
            else:
                status_code = 404
                response = {}

            requests_mock.get(api_url, [{"status_code": status_code, "json": response}])

            # anonymous
            assert get_canonical_github_origin_url(html_input_url) == expected_url

            # with credentials
            assert (
                get_canonical_github_origin_url(
                    html_input_url, credentials=github_credentials
                )
                == expected_url
            )

            # anonymous
            assert (
                GitHubSession(
                    user_agent="GitHub Session Test",
                ).get_canonical_url(html_input_url)
                == expected_url
            )

            # with credentials
            assert (
                GitHubSession(
                    user_agent="GitHub Session Test", credentials=github_credentials
                ).get_canonical_url(html_input_url)
                == expected_url
            )


def test_get_canonical_github_origin_url_not_gh_origin():
    """It should return the input url when that origin is not a github one"""
    url = "https://example.org"
    assert get_canonical_github_origin_url(url) == url

    assert (
        GitHubSession(
            user_agent="GitHub Session Test",
        ).get_canonical_url(url)
        == url
    )


def test_github_session_anonymous_session():
    user_agent = ("GitHub Session Test",)
    github_session = GitHubSession(
        user_agent=user_agent,
    )
    assert github_session.anonymous is True

    actual_headers = github_session.session.headers
    assert actual_headers["Accept"] == "application/vnd.github.v3+json"
    assert actual_headers["User-Agent"] == user_agent


@pytest.mark.parametrize(
    "num_ratelimit", [1]  # return a single rate-limit response, then continue
)
def test_github_session_ratelimit_once_recovery(
    caplog,
    mocker,
    github_requests_ratelimited,
    num_ratelimit,
    monkeypatch_sleep_calls,
    github_credentials,
):
    """GitHubSession should recover from hitting the rate-limit once"""
    caplog.set_level(logging.DEBUG, "swh.core.github.utils")

    github_session = GitHubSession(
        user_agent="GitHub Session Test", credentials=github_credentials
    )

    statsd_report = mocker.patch.object(github_session.statsd, "_report")

    res = github_session.request(f"{HTTP_GITHUB_API_URL}?per_page=1000&since=10")
    assert res.status_code == 200

    token_users = []
    for record in caplog.records:
        if "Using authentication token" in record.message:
            token_users.append(record.args[0])

    # check that we used one more token than we saw rate limited requests
    assert len(token_users) == 1 + num_ratelimit

    # check that we slept for one second between our token uses
    assert monkeypatch_sleep_calls == [1]

    username0 = github_session.credentials[0]["username"]
    username1 = github_session.credentials[1]["username"]
    tags0 = {"username": username0, "http_status": 403}
    tags1 = {"username": username1, "http_status": 200}
    assert [c for c in statsd_report.mock_calls] == [
        call("requests_total", "c", 1, {"username": username0}, 1),
        call("responses_total", "c", 1, tags0, 1),
        call("remaining_requests", "g", 999, {"username": username0}, 1),
        call("rate_limited_responses_total", "c", 1, {"username": username0}, 1),
        call("sleep_seconds_total", "c", 1, None, 1),
        call("requests_total", "c", 1, {"username": username1}, 1),
        call("responses_total", "c", 1, tags1, 1),
        call("remaining_requests", "g", 998, {"username": username1}, 1),
    ]
    assert github_session.statsd.constant_tags == {
        "api_type": "github",
        "api_instance": "github",
    }


def test_github_session_authenticated_credentials(
    caplog, github_credentials, all_tokens
):
    """GitHubSession should have Authorization headers set in authenticated mode"""
    caplog.set_level(logging.DEBUG, "swh.core.github.utils")

    github_session = GitHubSession(
        "GitHub Session Test", credentials=github_credentials
    )

    assert github_session.anonymous is False
    assert github_session.token_index == 0
    assert (
        sorted(github_session.credentials, key=lambda t: t["username"])
        == github_credentials
    )
    assert github_session.session.headers["Authorization"] in [
        f"token {t}" for t in all_tokens
    ]


@pytest.mark.parametrize(
    # Do 5 successful requests, return 6 ratelimits (to exhaust the credentials) with a
    # set value for X-Ratelimit-Reset, then resume listing successfully.
    "num_before_ratelimit, num_ratelimit, ratelimit_reset",
    [(5, 6, 123456)],
)
def test_github_session_ratelimit_reset_sleep(
    caplog,
    mocker,
    github_requests_ratelimited,
    monkeypatch_sleep_calls,
    num_before_ratelimit,
    num_ratelimit,
    ratelimit_reset,
    github_credentials,
):
    """GitHubSession should handle rate-limit with authentication tokens."""
    caplog.set_level(logging.DEBUG, "swh.core.github.utils")

    github_session = GitHubSession(
        user_agent="GitHub Session Test", credentials=github_credentials
    )

    statsd_report = mocker.patch.object(github_session.statsd, "_report")

    for _ in range(num_ratelimit):
        github_session.request(f"{HTTP_GITHUB_API_URL}?per_page=1000&since=10")

    # We sleep 1 second every time we change credentials, then we sleep until
    # ratelimit_reset + 1
    expected_sleep_calls = len(github_credentials) * [1] + [ratelimit_reset + 1]
    assert monkeypatch_sleep_calls == expected_sleep_calls

    found_exhaustion_message = False
    for record in caplog.records:
        if record.levelname == "INFO":
            if "Rate limits exhausted for all tokens" in record.message:
                found_exhaustion_message = True
                break

    assert found_exhaustion_message is True

    username0 = github_session.credentials[0]["username"]

    def ok_request_calls(user, remaining):
        return [
            call("requests_total", "c", 1, {"username": user}, 1),
            call("responses_total", "c", 1, {"username": user, "http_status": 200}, 1),
            call("remaining_requests", "g", remaining, {"username": user}, 1),
        ]

    def ratelimited_request_calls(user):
        return [
            call("requests_total", "c", 1, {"username": user}, 1),
            call("responses_total", "c", 1, {"username": user, "http_status": 403}, 1),
            call("remaining_requests", "g", 0, {"username": user}, 1),
            call("reset_seconds", "g", ratelimit_reset, {"username": user}, 1),
            call("rate_limited_responses_total", "c", 1, {"username": user}, 1),
            call("sleep_seconds_total", "c", 1, None, 1),
        ]

    expected_calls_groups = (
        # Successful requests
        [ok_request_calls(username0, n - 1) for n in range(num_before_ratelimit, 0, -1)]
        # Then rate-limited failures, cycling through tokens
        + [
            ratelimited_request_calls(
                github_session.credentials[n % len(github_credentials)]["username"]
            )
            for n in range(num_ratelimit)
        ]
        # And finally, a long sleep and the successful request
        + [
            [call("sleep_seconds_total", "c", ratelimit_reset + 1, None, 1)],
            ok_request_calls(
                github_session.credentials[num_ratelimit % len(github_credentials)][
                    "username"
                ],
                1000 - num_ratelimit - 1,
            ),
        ]
    )
    expected_calls = list(itertools.chain.from_iterable(expected_calls_groups))
    assert [c for c in statsd_report.mock_calls] == expected_calls
    assert github_session.statsd.constant_tags == {
        "api_type": "github",
        "api_instance": "github",
    }


# Same as before, but with no credentials
@pytest.mark.parametrize(
    "num_before_ratelimit, num_ratelimit, ratelimit_reset",
    [(5, 6, 123456)],
)
def test_github_session_ratelimit_reset_sleep_anonymous(
    caplog,
    mocker,
    github_requests_ratelimited,
    monkeypatch_sleep_calls,
    num_before_ratelimit,
    num_ratelimit,
    ratelimit_reset,
):
    """GitHubSession should handle rate-limit with authentication tokens."""
    caplog.set_level(logging.DEBUG, "swh.core.github.utils")

    github_session = GitHubSession(user_agent="GitHub Session Test")

    statsd_report = mocker.patch.object(github_session.statsd, "_report")

    for _ in range(num_ratelimit):
        github_session.request(f"{HTTP_GITHUB_API_URL}?per_page=1000&since=10")

    # No credentials, so we immediately sleep for a long time
    expected_sleep_calls = [ratelimit_reset + 1] * num_ratelimit
    assert monkeypatch_sleep_calls == expected_sleep_calls

    found_exhaustion_message = False
    for record in caplog.records:
        if record.levelname == "INFO":
            if "Rate limits exhausted for all tokens" in record.message:
                found_exhaustion_message = True
                break

    assert found_exhaustion_message is True

    user = "anonymous"

    def ok_request_calls(remaining):
        return [
            call("requests_total", "c", 1, {"username": user}, 1),
            call("responses_total", "c", 1, {"username": user, "http_status": 200}, 1),
            call("remaining_requests", "g", remaining, {"username": user}, 1),
        ]

    def ratelimited_request_calls():
        return [
            call("requests_total", "c", 1, {"username": user}, 1),
            call("responses_total", "c", 1, {"username": user, "http_status": 403}, 1),
            call("remaining_requests", "g", 0, {"username": user}, 1),
            call("reset_seconds", "g", ratelimit_reset, {"username": user}, 1),
            call("rate_limited_responses_total", "c", 1, {"username": user}, 1),
            call("sleep_seconds_total", "c", ratelimit_reset + 1, None, 1),
        ]

    expected_calls_groups = (
        # Successful requests
        [ok_request_calls(n - 1) for n in range(num_before_ratelimit, 0, -1)]
        # Then rate-limited failures, each with a long sleep
        + [ratelimited_request_calls() for n in range(num_ratelimit)]
        # And finally, the successful request
        + [
            ok_request_calls(
                1000 - num_ratelimit - 1,
            ),
        ]
    )
    expected_calls = list(itertools.chain.from_iterable(expected_calls_groups))
    assert [c for c in statsd_report.mock_calls] == expected_calls
    assert github_session.statsd.constant_tags == {
        "api_type": "github",
        "api_instance": "github",
    }


def test_github_session_get_repo_metadata_success(requests_mock):
    user_repo = KNOWN_GH_REPO.replace("https://github.com/", "")
    repo_metadata = {"html_url": KNOWN_GH_REPO}
    requests_mock.get(_url_github_api(user_repo), json=repo_metadata)

    gh_session = GitHubSession(user_agent="GitHub Session Test")
    assert gh_session.get_repository_metadata(KNOWN_GH_REPO) == repo_metadata


def test_github_session_get_repo_metadata_failure(requests_mock):
    unknown_user_repo = KNOWN_GH_REPO2.replace("https://github.com/", "")
    requests_mock.get(_url_github_api(unknown_user_repo), status_code=404)

    gh_session = GitHubSession(user_agent="GitHub Session Test")
    with pytest.raises(requests.HTTPError):
        gh_session.get_repository_metadata(KNOWN_GH_REPO2)



# Copyright (C) 2023-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Callable

from requests.exceptions import ChunkedEncodingError, ConnectionError, HTTPError
from requests.status_codes import codes
from tenacity import retry as tenacity_retry
from tenacity.stop import stop_after_attempt
from tenacity.wait import wait_exponential


def is_throttling_exception(e: Exception) -> bool:
    """
    Checks if an exception is a :class:`requests.exception.HTTPError` for
    a response with status code 429 (too many requests).
    """
    return (
        isinstance(e, HTTPError)
        and e.response is not None
        and e.response.status_code == codes.too_many_requests
    )


def is_retryable_exception(e: Exception) -> bool:
    """
    Checks if an exception is worth retrying (connection, throttling or a server error).
    """
    is_connection_error = isinstance(e, (ConnectionError, ChunkedEncodingError))
    is_500_error = (
        isinstance(e, HTTPError)
        and e.response is not None
        and e.response.status_code >= 500
    )

    return is_connection_error or is_throttling_exception(e) or is_500_error


def retry_if_exception(retry_state, predicate: Callable[[Exception], bool]) -> bool:
    """
    Custom tenacity retry predicate for handling exceptions with the given predicate.
    """
    attempt = retry_state.outcome
    if attempt.failed:
        exception = attempt.exception()
        return predicate(exception)
    return False


def retry_policy_generic(retry_state) -> bool:
    """
    Custom tenacity retry predicate for handling failed requests:
        - ConnectionError
        - Server errors (status >= 500)
        - Throttling errors (status == 429)

    This does not handle 403, 404 or other status codes.
    """
    return retry_if_exception(retry_state, is_retryable_exception)


WAIT_EXP_BASE = 10
MAX_NUMBER_ATTEMPTS = 5


def http_retry(
    retry=retry_policy_generic,
    wait=wait_exponential(exp_base=WAIT_EXP_BASE),
    stop=stop_after_attempt(max_attempt_number=MAX_NUMBER_ATTEMPTS),
    **retry_args,
):
    """
    Decorator based on :mod:`tenacity` for retrying a function possibly raising
    :class:`requests.exception.HTTPError` for status code 429 (too many requests)
    or >= 500.

    It provides a default configuration that should work properly in most
    cases but all :func:`tenacity.retry` parameters can also be overridden in client
    code.

    When the maximum of attempts is reached, the ``HTTPError`` exception will then
    be reraised.

    Args:
        retry: function defining request retry condition (default to 429 status code)
            https://tenacity.readthedocs.io/en/latest/#whether-to-retry

        wait: function defining wait strategy before retrying (default to exponential
            backoff) https://tenacity.readthedocs.io/en/latest/#waiting-before-retrying

        stop: function defining when to stop retrying (default after 5 attempts)
            https://tenacity.readthedocs.io/en/latest/#stopping

    """
    return tenacity_retry(retry=retry, wait=wait, stop=stop, reraise=True, **retry_args)



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import os

from click.testing import CliRunner
import pytest

from swh.core.pytest_plugin import requests_mock_datadir_factory
from swh.core.tests.test_cli import assert_result


def response_context_callback(request, context):
    """Add link headers to mocked Sentry REST API responses"""
    base_url = f"{request.scheme}://{request.netloc}{request.path}"
    if "cursor" not in request.query:
        context.headers["Link"] = f'<{base_url}?cursor=0:100:0>; rel="next"'
    else:
        context.headers["Link"] = f'<{base_url}?cursor=0:200:0>; rel="next"'


requests_mock_sentry = requests_mock_datadir_factory(
    response_context_callback=response_context_callback
)

SENTRY_ISSUE_ID = "112726"


@pytest.fixture
def swhmain(swhmain):
    from swh.core.cli.sentry import sentry as swhsentry

    swhmain.add_command(swhsentry)
    return swhmain


def test_sentry_extract_origin_urls(swhmain, requests_mock_sentry):
    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        ["sentry", "extract-origin-urls", "-t", "sentry-token", "-i", SENTRY_ISSUE_ID],
    )
    assert_result(result)
    expected_output = """
opam+https://opam.ocaml.org/packages/bdd/
opam+https://opam.ocaml.org/packages/bitv/
opam+https://opam.ocaml.org/packages/cgi/
opam+https://opam.ocaml.org/packages/combine/
"""
    assert result.output.strip() == expected_output.strip("\n")


def test_sentry_extract_scheduler_tasks(swhmain, requests_mock_sentry, datadir):
    runner = CliRunner()
    result = runner.invoke(
        swhmain,
        [
            "sentry",
            "extract-scheduler-tasks",
            "-t",
            "sentry-token",
            "-i",
            SENTRY_ISSUE_ID,
        ],
    )
    assert_result(result)
    csv_tasks_file = os.path.join(datadir, "sentry_expected_scheduler_tasks.csv")
    with open(csv_tasks_file, "r") as tasks_csv:
        assert result.output.strip() == tasks_csv.read()



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

import pytest
import sentry_sdk
from sentry_sdk import capture_exception, capture_message, set_tag

from swh.core.sentry import (
    init_sentry,
    override_with_bool_envvar,
    override_with_float_envvar,
)

SENTRY_DSN = "https://user@example.org/1234"


@pytest.mark.parametrize(
    "envvalue,retval",
    (
        ("y", True),
        ("n", False),
        ("0", False),
        ("true", True),
        ("FaLsE", False),
        ("1", True),
    ),
)
def test_override_with_bool_envvar(monkeypatch, envvalue: str, retval: bool):
    """Test if the override_with_bool_envvar function returns appropriate results"""
    envvar = "OVERRIDE_WITH_BOOL_ENVVAR"
    monkeypatch.setenv(envvar, envvalue)
    for default in (True, False):
        assert override_with_bool_envvar(envvar, default) == retval


def test_override_with_bool_envvar_logging(monkeypatch, caplog):
    envvar = "OVERRIDE_WITH_BOOL_ENVVAR"
    monkeypatch.setenv(envvar, "not a boolean env value")
    for default in (True, False):
        caplog.clear()
        assert override_with_bool_envvar(envvar, default) == default
        assert len(caplog.records) == 1
        assert (
            "OVERRIDE_WITH_BOOL_ENVVAR='not a boolean env value'"
            in caplog.records[0].getMessage()
        )
        assert f"using default value {default}" in caplog.records[0].getMessage()
        assert caplog.records[0].levelname == "WARNING"


@pytest.mark.parametrize(
    "envvalue,retval",
    (("1.0", 1.0), ("0.0", 0.0), ("0", 0.0), ("a", None)),
)
def test_override_with_float_envvar(monkeypatch, envvalue: str, retval: bool):
    envvar = "OVERRIDE_WITH_FLOAT_ENVVAR"
    monkeypatch.setenv(envvar, envvalue)
    assert override_with_float_envvar(envvar, None) == retval


def test_override_with_float_envvar_logging(monkeypatch, caplog):
    envvar = "OVERRIDE_WITH_FLOAT_ENVVAR"
    monkeypatch.setenv(envvar, "not a float env value")
    for default in (None, 1.0):
        caplog.clear()
        assert override_with_float_envvar(envvar, default) == default
        assert len(caplog.records) == 1
        assert (
            "OVERRIDE_WITH_FLOAT_ENVVAR='not a float env value'"
            in caplog.records[0].getMessage()
        )
        assert f"using default value {default}" in caplog.records[0].getMessage()
        assert caplog.records[0].levelname == "WARNING"


def test_sentry():
    reports = []
    init_sentry(SENTRY_DSN, extra_kwargs={"transport": reports.append})

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 2
    assert reports[0]["message"] == "Something went wrong"
    assert reports[1]["logentry"]["message"] == "Stupid error"


def test_sentry_no_logging():
    reports = []
    init_sentry(
        SENTRY_DSN,
        disable_logging_events=True,
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 1
    assert reports[0]["message"] == "Something went wrong"


def test_sentry_no_logging_from_venv(monkeypatch):
    monkeypatch.setenv("SWH_SENTRY_DISABLE_LOGGING_EVENTS", "True")

    reports = []
    init_sentry(
        SENTRY_DSN,
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 1
    assert reports[0]["message"] == "Something went wrong"


def test_sentry_logging_from_venv(monkeypatch):
    monkeypatch.setenv("SWH_SENTRY_DISABLE_LOGGING_EVENTS", "false")

    reports = []
    init_sentry(
        SENTRY_DSN,
        extra_kwargs={"transport": reports.append},
    )

    capture_message("Something went wrong")
    logging.error("Stupid error")

    assert len(reports) == 2


def test_sentry_events_fixture_capture_message(sentry_events):
    message = "Something went wrong"
    capture_message(message)
    assert sentry_events
    assert "message" in sentry_events[0]
    assert sentry_events[0]["message"] == message


def test_sentry_events_fixture_capture_exception(sentry_events):
    message = "Invalid value"
    exception = ValueError(message)
    capture_exception(exception)
    assert sentry_events
    assert "exception" in sentry_events[0]
    assert "values" in sentry_events[0]["exception"]
    exception_data = sentry_events[0]["exception"]["values"]
    assert exception_data
    assert exception_data[0].get("type") == type(exception).__name__
    assert exception_data[0].get("value") == message


def test_sentry_events_fixture_set_tag(sentry_events):
    tag_name = "swh.test"
    tag_value = "test"
    set_tag(tag_name, tag_value)
    message = "Something went wrong"
    capture_message(message)
    assert sentry_events
    assert "tags" in sentry_events[0]
    sentry_events[0]["tags"] == {tag_name: tag_value}


def test_sentry_main_package(mocker):
    sentry_sdk_init = mocker.patch.object(sentry_sdk, "init")
    init_sentry(
        SENTRY_DSN,
    )
    assert sentry_sdk_init.call_args_list[-1][1]["release"] is None

    init_sentry(SENTRY_DSN, main_package="swh.core")
    assert sentry_sdk_init.call_args_list[-1][1]["release"].startswith("swh.core@")


@pytest.mark.parametrize("deferred_init", [False, True])
def test_sentry_deferred_init(caplog, deferred_init):
    init_sentry(None, deferred_init=deferred_init)
    if not deferred_init:
        assert caplog.records
        assert (
            caplog.records[0].message
            == "Sentry DSN not provided, events will not be sent."
        )
    else:
        assert not caplog.records


@pytest.mark.parametrize("traces_sample_rate", [1.0, 0.0, None])
def test_sentry_traces_sample_rate(caplog, traces_sample_rate):
    init_sentry(None, traces_sample_rate=traces_sample_rate)
    client = sentry_sdk.get_client()
    if traces_sample_rate is not None:
        assert client.options["traces_sample_rate"] == traces_sample_rate
    else:
        assert client.options["enable_tracing"] is None
        assert client.options["traces_sample_rate"] is None



# Copyright (C) 2018-2021  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# Initially imported from https://github.com/DataDog/datadogpy/
# at revision 62b3a3e89988dc18d78c282fe3ff5d1813917436
#
# Copyright (c) 2015, Datadog <info@datadoghq.com>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of Datadog nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import socket
import time

import pytest

from swh.core.pytest_plugin import FakeSocket
from swh.core.statsd import Statsd, TimedContextManagerDecorator


class BrokenSocket(FakeSocket):
    def send(self, payload):
        raise socket.error("Socket error")


class SlowSocket(FakeSocket):
    def send(self, payload):
        raise socket.timeout("Socket timeout")


def assert_almost_equal(a, b, delta):
    assert 0 <= abs(a - b) <= delta, f"|{a} - {b}| not within {delta}"


def test_set(statsd):
    statsd.set("set", 123)
    assert statsd.socket.recv() == "set:123|s"


def test_gauge(statsd):
    statsd.gauge("gauge", 123.4)
    assert statsd.socket.recv() == "gauge:123.4|g"


def test_counter(statsd):
    statsd.increment("page.views")
    assert statsd.socket.recv() == "page.views:1|c"

    statsd.increment("page.views", 11)
    assert statsd.socket.recv() == "page.views:11|c"

    statsd.decrement("page.views")
    assert statsd.socket.recv() == "page.views:-1|c"

    statsd.decrement("page.views", 12)
    assert statsd.socket.recv() == "page.views:-12|c"


def test_histogram(statsd):
    statsd.histogram("histo", 123.4)
    assert statsd.socket.recv() == "histo:123.4|h"


def test_tagged_gauge(statsd):
    statsd.gauge("gt", 123.4, tags={"country": "china", "age": 45})
    assert statsd.socket.recv() == "gt:123.4|g|#age:45,country:china"


def test_tagged_counter(statsd):
    statsd.increment("ct", tags={"country": "españa"})
    assert statsd.socket.recv() == "ct:1|c|#country:españa"


def test_tagged_histogram(statsd):
    statsd.histogram("h", 1, tags={"test_tag": "tag_value"})
    assert statsd.socket.recv() == "h:1|h|#test_tag:tag_value"


def test_sample_rate(statsd):
    statsd.increment("c", sample_rate=0)
    assert not statsd.socket.recv()
    for i in range(10000):
        statsd.increment("sampled_counter", sample_rate=0.3)
    assert_almost_equal(3000, len(statsd.socket.payloads), 150)
    assert statsd.socket.recv() == "sampled_counter:1|c|@0.3"


def test_tags_and_samples(statsd):
    for i in range(100):
        statsd.gauge("gst", 23, tags={"sampled": True}, sample_rate=0.9)

    assert_almost_equal(90, len(statsd.socket.payloads), 10)
    assert statsd.socket.recv() == "gst:23|g|@0.9|#sampled:True"


def test_timing(statsd):
    statsd.timing("t", 123)
    assert statsd.socket.recv() == "t:123|ms"


def test_metric_namespace(statsd):
    """
    Namespace prefixes all metric names.
    """
    statsd.namespace = "foo"
    statsd.gauge("gauge", 123.4)
    assert statsd.socket.recv() == "foo.gauge:123.4|g"


# Test Client level constant tags
def test_gauge_constant_tags(statsd):
    statsd.constant_tags = {
        "bar": "baz",
    }
    statsd.gauge("gauge", 123.4)
    assert statsd.socket.recv() == "gauge:123.4|g|#bar:baz"


def test_counter_constant_tag_with_metric_level_tags(statsd):
    statsd.constant_tags = {
        "bar": "baz",
        "foo": True,
    }
    statsd.increment("page.views", tags={"extra": "extra"})
    assert statsd.socket.recv() == "page.views:1|c|#bar:baz,extra:extra,foo:True"


def test_gauge_constant_tags_with_metric_level_tags_twice(statsd):
    metric_level_tag = {"foo": "bar"}
    statsd.constant_tags = {"bar": "baz"}
    statsd.gauge("gauge", 123.4, tags=metric_level_tag)
    assert statsd.socket.recv() == "gauge:123.4|g|#bar:baz,foo:bar"

    # sending metrics multiple times with same metric-level tags
    # should not duplicate the tags being sent
    statsd.gauge("gauge", 123.4, tags=metric_level_tag)
    assert statsd.socket.recv() == "gauge:123.4|g|#bar:baz,foo:bar"


def test_socket_error(statsd):
    statsd._socket = BrokenSocket()
    statsd.gauge("no error", 1)
    assert True, "success"


def test_socket_timeout(statsd):
    statsd._socket = SlowSocket()
    statsd.gauge("no error", 1)
    assert True, "success"


def test_timed(statsd):
    """
    Measure the distribution of a function's run time.
    """

    @statsd.timed("timed.test")
    def func(a, b, c=1, d=1):
        """docstring"""
        time.sleep(0.5)
        return (a, b, c, d)

    assert func.__name__ == "func"
    assert func.__doc__ == "docstring"

    result = func(1, 2, d=3)
    # Assert it handles args and kwargs correctly.
    assert result, (1, 2, 1 == 3)

    packet = statsd.socket.recv()
    name_value, type_ = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "ms"
    assert name == "timed.test"
    assert_almost_equal(500, float(value), 100)


def test_timed_exception(statsd):
    """
    Exception bubble out of the decorator and is reported
    to statsd as a dedicated counter.
    """

    @statsd.timed("timed.test")
    def func(a, b, c=1, d=1):
        """docstring"""
        time.sleep(0.5)
        return (a / b, c, d)

    assert func.__name__ == "func"
    assert func.__doc__ == "docstring"

    with pytest.raises(ZeroDivisionError):
        func(1, 0)

    packet = statsd.socket.recv()
    name_value, type_, tags = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "c"
    assert name == "timed.test_error_count"
    assert int(value) == 1
    assert tags == "#error_type:ZeroDivisionError"


def test_timed_no_metric(statsd):
    """
    Test using a decorator without providing a metric.
    """

    @statsd.timed()
    def func(a, b, c=1, d=1):
        """docstring"""
        time.sleep(0.5)
        return (a, b, c, d)

    assert func.__name__ == "func"
    assert func.__doc__ == "docstring"

    result = func(1, 2, d=3)
    # Assert it handles args and kwargs correctly.
    assert result, (1, 2, 1 == 3)

    packet = statsd.socket.recv()
    name_value, type_ = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "ms"
    assert name in ("swh.core.tests.test_statsd.func", "core.tests.test_statsd.func")
    assert_almost_equal(500, float(value), 100)


def test_timed_coroutine(statsd):
    """
    Measure the distribution of a coroutine function's run time.
    """
    import asyncio

    @statsd.timed("timed.test")
    async def print_foo():
        """docstring"""
        time.sleep(0.5)
        print("foo")

    loop = asyncio.new_event_loop()
    loop.run_until_complete(print_foo())
    loop.close()

    # Assert
    packet = statsd.socket.recv()
    name_value, type_ = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "ms"
    assert name == "timed.test"
    assert_almost_equal(500, float(value), 100)


def test_timed_context(statsd):
    """
    Measure the distribution of a context's run time.
    """
    # In milliseconds
    with statsd.timed("timed_context.test") as timer:
        assert isinstance(timer, TimedContextManagerDecorator)
        time.sleep(0.5)

    packet = statsd.socket.recv()
    name_value, type_ = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "ms"
    assert name == "timed_context.test"
    assert_almost_equal(500, float(value), 100)
    assert_almost_equal(500, timer.elapsed, 100)


def test_timed_context_exception(statsd):
    """
    Exception bubbles out of the `timed` context manager and is
    reported to statsd as a dedicated counter.
    """

    class ContextException(Exception):
        pass

    def func(statsd):
        with statsd.timed("timed_context.test"):
            time.sleep(0.5)
            raise ContextException()

    # Ensure the exception was raised.
    with pytest.raises(ContextException):
        func(statsd)

    # Ensure the timing was recorded.
    packet = statsd.socket.recv()
    name_value, type_, tags = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "c"
    assert name == "timed_context.test_error_count"
    assert int(value) == 1
    assert tags == "#error_type:ContextException"


def test_timed_context_no_metric_name_exception(statsd):
    """Test that an exception occurs if using a context manager without a
    metric name.
    """

    def func(statsd):
        with statsd.timed():
            time.sleep(0.5)

    # Ensure the exception was raised.
    with pytest.raises(TypeError):
        func(statsd)

    # Ensure the timing was recorded.
    packet = statsd.socket.recv()
    assert packet is None


def test_timed_start_stop_calls(statsd):
    timer = statsd.timed("timed_context.test")
    timer.start()
    time.sleep(0.5)
    timer.stop()

    packet = statsd.socket.recv()
    name_value, type_ = packet.split("|")
    name, value = name_value.split(":")

    assert type_ == "ms"
    assert name == "timed_context.test"
    assert_almost_equal(500, float(value), 100)


def test_batched(statsd):
    statsd.open_buffer()
    statsd.gauge("page.views", 123)
    statsd.timing("timer", 123)
    statsd.close_buffer()

    assert statsd.socket.recv() == "page.views:123|g\ntimer:123|ms"


def test_context_manager():
    fake_socket = FakeSocket()
    with Statsd() as statsd:
        statsd._socket = fake_socket
        statsd.gauge("page.views", 123)
        statsd.timing("timer", 123)

    assert fake_socket.recv() == "page.views:123|g\ntimer:123|ms"


def test_batched_buffer_autoflush():
    fake_socket = FakeSocket()
    with Statsd() as statsd:
        statsd._socket = fake_socket
        for i in range(51):
            statsd.increment("mycounter")
        assert "\n".join(["mycounter:1|c" for i in range(50)]) == fake_socket.recv()

    assert fake_socket.recv() == "mycounter:1|c"


def test_module_level_instance(statsd):
    from swh.core.statsd import statsd

    assert isinstance(statsd, Statsd)


def test_instantiating_does_not_connect():
    local_statsd = Statsd()
    assert local_statsd._socket is None


def test_accessing_socket_opens_socket():
    local_statsd = Statsd()
    try:
        assert local_statsd.socket is not None
    finally:
        local_statsd.close_socket()


def test_accessing_socket_multiple_times_returns_same_socket():
    local_statsd = Statsd()
    fresh_socket = FakeSocket()
    local_statsd._socket = fresh_socket
    assert fresh_socket == local_statsd.socket
    assert FakeSocket() != local_statsd.socket


def test_tags_from_environment(monkeypatch):
    monkeypatch.setenv("STATSD_TAGS", "country:china,age:45")
    statsd = Statsd()
    statsd._socket = FakeSocket()
    statsd.gauge("gt", 123.4)
    assert statsd.socket.recv() == "gt:123.4|g|#age:45,country:china"


def test_tags_from_environment_with_substitution(monkeypatch):
    monkeypatch.setenv("HOSTNAME", "sweethome")
    monkeypatch.setenv("PORT", "42")
    monkeypatch.setenv(
        "STATSD_TAGS", "country:china,age:45,host:$HOSTNAME,port:${PORT}"
    )
    statsd = Statsd()
    statsd._socket = FakeSocket()
    statsd.gauge("gt", 123.4)
    assert (
        statsd.socket.recv()
        == "gt:123.4|g|#age:45,country:china,host:sweethome,port:42"
    )


def test_tags_from_environment_and_constant(monkeypatch):
    monkeypatch.setenv("STATSD_TAGS", "country:china,age:45")
    statsd = Statsd(constant_tags={"country": "canada"})
    statsd._socket = FakeSocket()
    statsd.gauge("gt", 123.4)
    assert statsd.socket.recv() == "gt:123.4|g|#age:45,country:canada"


def test_tags_from_environment_warning(monkeypatch):
    monkeypatch.setenv("STATSD_TAGS", "valid:tag,invalid_tag")
    with pytest.warns(UserWarning) as record:
        statsd = Statsd()

    assert len(record) == 1
    assert "invalid_tag" in record[0].message.args[0]
    assert "valid:tag" not in record[0].message.args[0]
    assert statsd.constant_tags == {"valid": "tag"}


def test_gauge_doesnt_send_none(statsd):
    statsd.gauge("metric", None)
    assert statsd.socket.recv() is None


def test_increment_doesnt_send_none(statsd):
    statsd.increment("metric", None)
    assert statsd.socket.recv() is None


def test_decrement_doesnt_send_none(statsd):
    statsd.decrement("metric", None)
    assert statsd.socket.recv() is None


def test_timing_doesnt_send_none(statsd):
    statsd.timing("metric", None)
    assert statsd.socket.recv() is None


def test_histogram_doesnt_send_none(statsd):
    statsd.histogram("metric", None)
    assert statsd.socket.recv() is None


def test_param_host(monkeypatch):
    monkeypatch.setenv("STATSD_HOST", "test-value")
    monkeypatch.setenv("STATSD_PORT", "")
    local_statsd = Statsd(host="actual-test-value")

    assert local_statsd.host == "actual-test-value"
    assert local_statsd.port == 8125


def test_param_port(monkeypatch):
    monkeypatch.setenv("STATSD_HOST", "")
    monkeypatch.setenv("STATSD_PORT", "12345")
    local_statsd = Statsd(port=4321)
    assert local_statsd.host == "localhost"
    assert local_statsd.port == 4321


def test_envvar_host(monkeypatch):
    monkeypatch.setenv("STATSD_HOST", "test-value")
    monkeypatch.setenv("STATSD_PORT", "")
    local_statsd = Statsd()
    assert local_statsd.host == "test-value"
    assert local_statsd.port == 8125


def test_envvar_port(monkeypatch):
    monkeypatch.setenv("STATSD_HOST", "")
    monkeypatch.setenv("STATSD_PORT", "12345")
    local_statsd = Statsd()

    assert local_statsd.host == "localhost"
    assert local_statsd.port == 12345


def test_namespace_added():
    local_statsd = Statsd(namespace="test-namespace")
    local_statsd._socket = FakeSocket()

    local_statsd.gauge("gauge", 123.4)
    assert local_statsd.socket.recv() == "test-namespace.gauge:123.4|g"


def test_contextmanager_empty(statsd):
    with statsd:
        assert True, "success"


def test_contextmanager_buffering(statsd):
    with statsd as s:
        s.gauge("gauge", 123.4)
        s.gauge("gauge_other", 456.78)
        assert s.socket.recv() is None

    assert statsd.socket.recv() == "gauge:123.4|g\ngauge_other:456.78|g"


def test_timed_elapsed(statsd):
    with statsd.timed("test_timer") as t:
        pass

    assert t.elapsed >= 0
    assert statsd.socket.recv() == "test_timer:%s|ms" % t.elapsed


def test_status_gauge(statsd):
    with statsd.status_gauge("test_status_gauge", ["s1", "s2", "s3"]) as set_status:
        set_status("s1")
        set_status("s2")
        set_status("s3")

    # enter the context manager: initialisation of gauges for listed statuses
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s2"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s3"
    # set_status("s1")
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s1"
    # set_status("s2")
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s2"
    # set_status("s3")
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s2"
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s3"
    # exit the context manager: cleanup gauges
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s2"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s3"


def test_status_gauge_error_on_exit(statsd):
    with pytest.raises(Exception):
        with statsd.status_gauge("test_status_gauge", ["s1"]) as set_status:
            set_status("s1")
            raise Exception()
    # enter the context manager: initialisation of gauges for listed statuses
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    # set_status("s1")
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s1"
    # exit the context manager: cleanup gauges
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"


def test_status_gauge_error(statsd):
    with statsd.status_gauge("test_status_gauge", ["s1", "s2", "s3"]) as set_status:
        with pytest.raises(ValueError):
            set_status("s4")


def test_status_gauge_repeated(statsd):
    with statsd.status_gauge("test_status_gauge", ["s1", "s2", "s3"]) as set_status:
        set_status("s1")
        set_status("s1")
        set_status("s1")

    # enter the context manager: initialisation of gauges for listed statuses
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s2"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s3"
    # set_status("s1")
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s1"
    # set_status("s1")
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s1"
    # set_status("s1")
    assert statsd.socket.recv() == "test_status_gauge:1|g|#status:s1"
    # exit the context manager: cleanup gauges
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s1"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s2"
    assert statsd.socket.recv() == "test_status_gauge:0|g|#status:s3"



# Copyright (C) 2018-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

import pytest
from requests.exceptions import ConnectionError

from swh.core.api import (
    RETRY_WAIT_INTERVAL,
    APIError,
    RemoteException,
    RPCClient,
    TransientRemoteException,
    remote_api_endpoint,
)
from swh.core.api.serializers import exception_to_dict, msgpack_dumps
from swh.core.retry import MAX_NUMBER_ATTEMPTS

from .test_serializers import ExtraType, extra_decoders, extra_encoders


class ReraiseException(Exception):
    pass


@pytest.fixture
def rpc_client_class(requests_mock):
    class TestStorage:
        @remote_api_endpoint("test_endpoint_url")
        def test_endpoint(self, test_data, db=None, cur=None): ...

        @remote_api_endpoint("path/to/endpoint")
        def something(self, data, db=None, cur=None): ...

        @remote_api_endpoint("serializer_test")
        def serializer_test(self, data, db=None, cur=None): ...

        @remote_api_endpoint("overridden/endpoint")
        def overridden_method(self, data):
            return "foo"

        @remote_api_endpoint("request_too_large")
        def request_too_large(self, data, db=None, cur=None): ...

    class Testclient(RPCClient):
        backend_class = TestStorage
        extra_type_encoders = extra_encoders
        extra_type_decoders = extra_decoders
        reraise_exceptions = [ReraiseException]
        enable_requests_retry = True

        def overridden_method(self, data):
            return "bar"

    def callback(request, context):
        assert request.headers["Content-Type"] == "application/x-msgpack"
        context.headers["Content-Type"] = "application/x-msgpack"
        if request.path == "/test_endpoint_url":
            context.content = b"\xa3egg"
        elif request.path == "/path/to/endpoint":
            context.content = b"\xa4spam"
        elif request.path == "/serializer_test":
            context.content = (
                b"\x82\xc4\x07swhtype\xa9extratype"
                b"\xc4\x01d\x92\x81\xa4spam\xa3egg\xa3qux"
            )
        elif request.path == "/request_too_large":
            context.status_code = 413
            context.headers["Content-Type"] = "text/html"
            context.content = b"<h1>413 request entity too large</h1>\r\n"
        else:
            assert False
        return context.content

    requests_mock.post(re.compile("mock://example.com/"), content=callback)

    return Testclient


@pytest.fixture
def rpc_client(rpc_client_class):
    return rpc_client_class(url="mock://example.com")


def test_client(rpc_client):
    assert hasattr(rpc_client, "test_endpoint")
    assert hasattr(rpc_client, "something")

    res = rpc_client.test_endpoint("spam")
    assert res == "egg"
    res = rpc_client.test_endpoint(test_data="spam")
    assert res == "egg"

    res = rpc_client.something("whatever")
    assert res == "spam"
    res = rpc_client.something(data="whatever")
    assert res == "spam"


def test_client_extra_serializers(rpc_client):
    res = rpc_client.serializer_test(["foo", ExtraType("bar", b"baz")])
    assert res == ExtraType({"spam": "egg"}, "qux")


def test_client_overridden_method(rpc_client):
    res = rpc_client.overridden_method("foo")
    assert res == "bar"


def test_client_request_too_large(rpc_client):
    with pytest.raises(APIError) as exc_info:
        rpc_client.request_too_large("foo")

    assert exc_info.value.args[0] == "<h1>413 request entity too large</h1>\r\n"
    assert exc_info.value.args[1].status_code == 413


def test_client_connexion_error(rpc_client, requests_mock, mocker):
    """
    ConnectionError should be wrapped and raised as an APIError.
    """
    mock_sleep = mocker.patch("time.sleep")
    error_message = "unreachable host"
    requests_mock.post(
        re.compile("mock://example.com/connection_error"),
        exc=ConnectionError(error_message),
    )

    with pytest.raises(APIError) as exc_info:
        rpc_client._post("connection_error", data={})

    assert type(exc_info.value.args[0]) is ConnectionError
    assert str(exc_info.value.args[0]) == error_message

    # check request retries on connection errors
    mock_sleep.assert_has_calls(
        [
            mocker.call(param)
            for param in [RETRY_WAIT_INTERVAL] * (MAX_NUMBER_ATTEMPTS - 1)
        ]
    )


def _exception_response(exception, status_code):
    def callback(request, context):
        assert request.headers["Content-Type"] == "application/x-msgpack"
        context.headers["Content-Type"] = "application/x-msgpack"
        exc_dict = exception_to_dict(exception)
        context.content = msgpack_dumps(exc_dict)
        context.status_code = status_code
        return context.content

    return callback


def test_client_reraise_exception(rpc_client, requests_mock, mocker):
    """
    Exception caught server-side and whitelisted will be raised again client-side.
    """
    mock_sleep = mocker.patch("time.sleep")
    error_message = "something went wrong"
    endpoint = "reraise_exception"

    requests_mock.post(
        re.compile(f"mock://example.com/{endpoint}"),
        content=_exception_response(
            exception=ReraiseException(error_message),
            status_code=400,
        ),
    )

    with pytest.raises(ReraiseException) as exc_info:
        rpc_client._post(endpoint, data={})

    assert str(exc_info.value) == error_message
    # no request retry for such exception
    mock_sleep.assert_not_called()


@pytest.mark.parametrize("status_code", [400, 500, 502, 503])
def test_client_raise_remote_exception(rpc_client, requests_mock, status_code, mocker):
    """
    Exception caught server-side and not whitelisted will be wrapped and raised
    as a RemoteException client-side.
    """
    error_message = "something went wrong"
    endpoint = "raise_remote_exception"

    requests_mock.post(
        re.compile(f"mock://example.com/{endpoint}"),
        content=_exception_response(
            exception=Exception(error_message),
            status_code=status_code,
        ),
    )
    mock_sleep = mocker.patch("time.sleep")

    with pytest.raises(RemoteException) as exc_info:
        rpc_client._post(endpoint, data={})

    assert str(exc_info.value.args[0]["type"]) == "Exception"
    assert str(exc_info.value.args[0]["message"]) == error_message
    if status_code in (502, 503):
        assert isinstance(exc_info.value, TransientRemoteException)
        # check request retry on transient remote exception
        mock_sleep.assert_has_calls(
            [
                mocker.call(param)
                for param in [RETRY_WAIT_INTERVAL] * (MAX_NUMBER_ATTEMPTS - 1)
            ]
        )
    else:
        assert not isinstance(exc_info.value, TransientRemoteException)
        # no request retry on other remote exceptions
        mock_sleep.assert_not_called()


@pytest.mark.parametrize(
    "timeout_arg,timeout_value",
    [
        pytest.param(None, None, id="default"),
        pytest.param(1.0, 1.0, id="float"),
        pytest.param((1, 2), (1, 2), id="tuple"),
        pytest.param([1, 2], (1, 2), id="list"),
    ],
)
def test_client_timeout_param(rpc_client_class, timeout_arg, timeout_value):
    client = rpc_client_class(url="mock://example.com/", timeout=timeout_arg)
    assert client.timeout == timeout_value


def test_client_timeout_valueerror(rpc_client_class):
    for timeout in ([], [1], [1, 2, 3]):
        with pytest.raises(ValueError) as exc:
            rpc_client_class(url="mock://example.com/", timeout=timeout)
        assert repr(timeout) in str(exc.value)



# Copyright (C) 2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from datetime import timedelta
from os import path

from psycopg2.errors import InsufficientPrivilege
import pytest

from swh.core.cli.db import db as swhdb
from swh.core.db import BaseDb
from swh.core.db.db_utils import (
    swh_db_module,
    swh_db_upgrade,
    swh_db_version,
    swh_db_versions,
    swh_set_db_module,
)
from swh.core.db.db_utils import get_database_info, get_sql_for_package, now
from swh.core.db.db_utils import parse_dsn_or_dbname as parse_dsn
from swh.core.tests.test_cli import assert_result

from .test_cli import craft_conninfo


def test_get_sql_for_package(mock_import_module):
    module = "test.postgresql"

    files = get_sql_for_package(module)
    assert files
    assert [f.name for f in files] == [
        "0-superuser-init.sql",
        "15-flavor.sql",
        "30-schema.sql",
        "40-funcs.sql",
        "50-data.sql",
    ]


def test_db_utils_versions(cli_runner, postgresql, mock_get_entry_points):
    """Check get_database_info, swh_db_versions and swh_db_module work ok

    This test checks db versions is properly initialized by the cli db init
    script.

    mock_import_swhmodule should set the initial version to 3.
    """
    module = "test"
    db_module = "test:postgresql"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert_result(result)

    # check the swh_db_module() function
    assert swh_db_module(conninfo) == db_module

    # the dbversion and dbmodule tables exists and are populated
    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    # check also the swh_db_versions() function
    versions = swh_db_versions(conninfo)

    assert dbmodule == db_module
    assert dbversion == 3
    assert dbflavor == "default"
    # check also the swh_db_versions() function
    versions = swh_db_versions(conninfo)
    assert len(versions) == 1
    assert versions[0][0] == 3
    assert versions[0][2] == "DB initialization"

    # add a few versions in dbversion
    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.executemany(
            "insert into dbversion(version, release, description) values (%s, %s, %s)",
            [(i, now(), f"Upgrade to version {i}") for i in range(4, 8)],
        )

    dbmodule, dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == db_module
    assert dbversion == 7
    assert dbflavor == "default"

    versions = swh_db_versions(conninfo)
    assert len(versions) == 5

    for i, (version, ts, desc) in enumerate(versions):
        assert version == (7 - i)  # these are in reverse order
        if version > 3:
            assert desc == f"Upgrade to version {version}"
            assert (now() - ts) < timedelta(seconds=1)


def test_db_utils_upgrade(cli_runner, postgresql, mock_get_entry_points, datadir):
    """Check swh_db_upgrade"""
    module = "test"
    db_module = "test:postgresql"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert_result(result)

    assert swh_db_version(conninfo) == 3
    new_version = swh_db_upgrade(conninfo, db_module)
    assert new_version == 6
    assert swh_db_version(conninfo) == 6

    versions = swh_db_versions(conninfo)
    # get rid of dates to ease checking
    versions = [(v[0], v[2]) for v in versions]
    assert versions[-1] == (3, "DB initialization")
    sqlbasedir = path.join(datadir, "test", "postgresql", "sql", "upgrades")

    assert versions[1:-1] == [
        (i, f"Upgraded to version {i} using {sqlbasedir}/{i:03d}.sql")
        for i in range(5, 3, -1)
    ]
    assert versions[0] == (6, "Updated version from upgrade script")

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("select url from origin where url like 'version%'")
        result = cur.fetchall()

        assert result == [("version%03d" % i,) for i in range(4, 7)]
        cur.execute(
            "select url from origin where url = 'this should never be executed'"
        )
        result = cur.fetchall()
        assert not result


def test_db_utils_swh_db_upgrade_sanity_checks(
    cli_runner, postgresql, mock_get_entry_points, datadir
):
    """Check swh_db_upgrade"""
    module = "test"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert_result(result)
    result = cli_runner.invoke(swhdb, ["init", module, "--dbname", conninfo])
    assert_result(result)

    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("drop table dbmodule")

    # try to upgrade with a unset module
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)

    # check the dbmodule is unset
    assert swh_db_module(conninfo) is None

    # set the stored module to something else
    swh_set_db_module(conninfo, f"{module}2")
    assert swh_db_module(conninfo) == f"{module}2"

    # try to upgrade with a different module
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)

    # revert to the proper module in the db
    swh_set_db_module(conninfo, module, force=True)
    assert swh_db_module(conninfo) == module
    # trying again is a noop
    swh_set_db_module(conninfo, module)
    assert swh_db_module(conninfo) == module

    # drop the dbversion table
    with cnx.transaction() as cur:
        cur.execute("drop table dbversion")
    # an upgrade should fail due to missing stored version
    with pytest.raises(ValueError):
        swh_db_upgrade(conninfo, module)


@pytest.mark.parametrize("flavor", [None, "default", "flavorA", "flavorB"])
def test_db_utils_flavor(cli_runner, postgresql, mock_get_entry_points, flavor):
    """Check populate_database_for_package handle db flavor properly"""
    module = "test"
    db_module = "test:postgresql"
    conninfo = craft_conninfo(postgresql)
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert_result(result)
    cmd = ["init", module, "--dbname", conninfo]
    if flavor:
        cmd.extend(["--flavor", flavor])
    result = cli_runner.invoke(swhdb, cmd)
    assert_result(result)

    # check the swh_db_module() function
    assert swh_db_module(conninfo) == db_module

    # the dbversion and dbmodule tables exists and are populated
    dbmodule, _dbversion, dbflavor = get_database_info(conninfo)
    assert dbmodule == db_module
    assert dbflavor == (flavor or "default")


def test_db_utils_guest_permissions(cli_runner, postgresql, mock_get_entry_points):
    """Check populate_database_for_package handle db flavor properly"""
    module = "test"
    conninfo = craft_conninfo(postgresql)
    # breakpoint()
    result = cli_runner.invoke(swhdb, ["init-admin", module, "--dbname", conninfo])
    assert_result(result)
    cmd = ["init", module, "--dbname", conninfo]
    result = cli_runner.invoke(swhdb, cmd)
    assert_result(result)

    # check select permissions have been granted tguest
    cnx = BaseDb.connect(conninfo)
    with cnx.transaction() as cur:
        cur.execute("select * from pg_roles where rolname='guest'")
        assert cur.rowcount == 1
        query = """
SELECT table_name, privilege_type
FROM information_schema.role_table_grants
WHERE grantee = 'guest'
"""
        cur.execute(query)
        assert cur.rowcount == 4
        assert set(cur.fetchall()) == {
            ("dbflavor", "SELECT"),
            ("origin", "SELECT"),
            ("dbversion", "SELECT"),
            ("dbmodule", "SELECT"),
        }

    guest_dsn = {**parse_dsn(conninfo), **{"user": "guest", "password": "guest"}}
    gcnx = BaseDb.connect(**guest_dsn)
    # check guest user can actually query a table
    with gcnx.transaction() as gcur:
        gcur.execute("select * from dbversion limit 1")
        assert gcur.rowcount == 1

    with gcnx.transaction() as gcur:
        # check guest user CANNOT create a new table
        with pytest.raises(InsufficientPrivilege):
            gcur.execute("create table toto(id int)")
    with gcnx.transaction() as gcur:
        # check guest user CANNOT drop a new table
        with pytest.raises(InsufficientPrivilege):
            gcur.execute("drop table origin")
    with gcnx.transaction() as gcur:
        # check guest user CANNOT insert data in a table
        with pytest.raises(InsufficientPrivilege):
            gcur.execute(
                """
INSERT INTO origin(url, hash)
VALUES ('https://nowhere.com', hash_sha1('https://nowhere.com'))
            """
            )



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""Logging module providing common swh logging configuration. This only depends on
python core library.

"""

import logging
from logging.config import dictConfig
from pathlib import Path
from typing import List, Optional, Tuple

from yaml import safe_load


class BadLogLevel(ValueError):
    pass


def validate_loglevel(value: str) -> Tuple[Optional[str], int]:
    """Validate a single loglevel specification, of the form LOGLEVEL or
    module:LOGLEVEL."""

    LOG_LEVEL_NAMES = ["notset", "debug", "info", "warning", "error", "critical"]

    if ":" in value:
        try:
            module, log_level = value.split(":")
        except ValueError:
            raise BadLogLevel(
                "Invalid log level specification `%s`, "
                "needs to be in format `module:LOGLEVEL`" % value
            )
    else:
        module = None
        log_level = value

    if log_level.lower() not in LOG_LEVEL_NAMES:
        raise BadLogLevel(
            f"Log level {log_level} unknown (in `{value}`) needs to be one "
            f"of {', '.join(LOG_LEVEL_NAMES)}"
        )

    return (module, logging.getLevelName(log_level.upper()))


def logging_configure(
    log_levels: List[Tuple[str, int]] = [], log_config: Optional[Path] = None
) -> str:
    """A default configuration function to unify swh module logger configuration.

    The log_config YAML file must conform to the logging.config.dictConfig schema
    documented at https://docs.python.org/3/library/logging.config.html.

    Returns:
        The actual root logger log level name defined.

    """
    set_default_loglevel: Optional[str] = None

    if log_config:
        with open(log_config, "r") as f:
            config_dict = safe_load(f.read())
        # Configure logging using a dictionary config
        dictConfig(config_dict)
        effective_level = logging.root.getEffectiveLevel()
        set_default_loglevel = logging.getLevelName(effective_level)

    if not log_levels:
        log_levels = []

    for module, log_level in log_levels:
        logger = logging.getLogger(module)
        logger.setLevel(log_level)

        if module is None:
            set_default_loglevel = log_level

    if not set_default_loglevel:
        logging.root.setLevel("INFO")
        set_default_loglevel = "INFO"

    return set_default_loglevel



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


from swh.core.api import remote_api_endpoint


def test_remote_api_endpoint():
    @remote_api_endpoint("hello_route")
    def hello():
        pass

    assert hasattr(hello, "_endpoint_path")
    assert hello._endpoint_path == "hello_route"
    assert hasattr(hello, "_method")
    assert hello._method == "POST"


def test_remote_api_endpoint_2():
    @remote_api_endpoint("another_route", method="GET")
    def hello2():
        pass

    assert hasattr(hello2, "_endpoint_path")
    assert hello2._endpoint_path == "another_route"
    assert hasattr(hello2, "_method")
    assert hello2._method == "GET"



#!/usr/bin/env python3
# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Callable, Dict, List

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group


def common_options(func):
    import functools

    @click.option(
        "--sentry-url",
        "-u",
        default="https://sentry.softwareheritage.org",
        show_default=True,
        help="Sentry URL",
    )
    @click.option(
        "--sentry-token",
        "-t",
        default=None,
        envvar="SENTRY_TOKEN",
        help=(
            "Bearer token required to communicate with Sentry API (can also be provided "
            "in SENTRY_TOKEN environment variable)"
        ),
        required=True,
    )
    @click.option(
        "--sentry-issue-number",
        "-i",
        help="Sentry issue number to extract origin URLs from its events",
        required=True,
    )
    @click.option(
        "--environment",
        "-e",
        default="",
        help="Filter on environment: production or staging, both are selected by default",
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@swh_cli_group.group(name="sentry", context_settings=CONTEXT_SETTINGS)
def sentry():
    """Software Heritage tools for extracting data from the events associated to
    a Sentry issue using Sentry REST API."""
    pass


def _process_sentry_events_pages(
    sentry_url,
    sentry_token,
    sentry_issue_number,
    events_page_process_callback: Callable[[List[Dict[str, Any]]], None],
    full_sentry_responses: bool = False,
):
    import requests

    sentry_api_base_url = f"{sentry_url.rstrip('/')}/api/0"
    sentry_issue_events_url = (
        f"{sentry_api_base_url}/issues/{sentry_issue_number}/events/"
    )
    if full_sentry_responses:
        sentry_issue_events_url += "?full=true"
    while True:
        response = requests.get(
            sentry_issue_events_url, headers={"Authorization": f"Bearer {sentry_token}"}
        )
        events = response.json()
        if not events:
            break
        events_page_process_callback(events)
        sentry_issue_events_url = response.links.get("next", {}).get("url")


@sentry.command(name="extract-origin-urls", context_settings=CONTEXT_SETTINGS)
@common_options
def extract_origin_urls(sentry_url, sentry_token, sentry_issue_number, environment):
    """Extract origin URLs from events.

    This command allows to extract origin URLs from Sentry events related to
    a Software Heritage loader and dumps them to stdout."""

    origin_urls = set()

    def _extract_origin_urls(events: List[Dict[str, Any]]):
        for event in events:
            tags = {tag["key"]: tag["value"] for tag in event.get("tags", [])}
            env_match = environment in tags.get("environment", "")
            if "swh.loader.origin_url" in tags and env_match:
                origin_urls.add(tags["swh.loader.origin_url"])

    _process_sentry_events_pages(
        sentry_url,
        sentry_token,
        sentry_issue_number,
        _extract_origin_urls,
    )

    for origin_url in sorted(origin_urls):
        click.echo(origin_url)


@sentry.command(name="extract-scheduler-tasks", context_settings=CONTEXT_SETTINGS)
@common_options
def extract_scheduler_tasks(sentry_url, sentry_token, sentry_issue_number, environment):
    """Extract scheduler task parameters from events.

    This command allows to extract scheduler task parameters from Sentry events related to
    a Software Heritage scheduler task and dumps a CSV file to stdout that can be consumed
    by the CLI command:

    $ swh scheduler task schedule --columns type --columns kwargs <csv_file>.
    """
    import csv
    import json
    import sys

    task_params = {}

    def _extract_scheduler_tasks(events):
        for event in events:
            celery_job = event.get("context", {}).get("celery-job", {})
            task_name = celery_job.get("task_name")
            task_param = celery_job.get("kwargs")
            if task_param:
                key = tuple([task_name] + list(task_param.values()))
                task_params[key] = (task_name, task_param)

    _process_sentry_events_pages(
        sentry_url,
        sentry_token,
        sentry_issue_number,
        _extract_scheduler_tasks,
        full_sentry_responses=True,
    )

    csv_writer = csv.writer(sys.stdout)
    for task_type, task_param in sorted(
        task_params.values(), key=lambda p: p[1].get("url", "")
    ):
        csv_writer.writerow([task_type, json.dumps(task_param)])



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import hashlib
import io
import os
import shutil
import stat
import tarfile

import pytest

from swh.core import tarball


@pytest.fixture
def prepare_shutil_state():
    """Reset any shutil modification in its current state"""
    import shutil

    registered_formats = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        name = format_id[0]
        if name in registered_formats:
            shutil.unregister_unpack_format(name)

    return shutil


def test_compress_uncompress_zip(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ("file%s.txt" % i)
        fpath.write_text("content of file %s" % i)

    zipfile = tmp_path / "archive.zip"
    tarball.compress(str(zipfile), "zip", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(zipfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ["file%s.txt" % i for i in range(10)] == lsdir


@pytest.mark.xfail(
    reason=(
        "Python's zipfile library doesn't support Info-ZIP's "
        "extension for file permissions."
    )
)
def test_compress_uncompress_zip_modes(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    fpath = tocompress / "text.txt"
    fpath.write_text("echo foo")
    fpath.chmod(0o644)

    fpath = tocompress / "executable.sh"
    fpath.write_text("echo foo")
    fpath.chmod(0o755)

    zipfile = tmp_path / "archive.zip"
    tarball.compress(str(zipfile), "zip", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(zipfile), str(destdir))

    (executable_path, text_path) = sorted(destdir.iterdir())
    assert text_path.stat().st_mode == 0o100644  # succeeds, it's the default
    assert executable_path.stat().st_mode == 0o100755  # fails


def test_compress_uncompress_tar(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    for i in range(10):
        fpath = tocompress / ("file%s.txt" % i)
        fpath.write_text("content of file %s" % i)

    tarfile = tmp_path / "archive.tar"
    tarball.compress(str(tarfile), "tar", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(tarfile), str(destdir))

    lsdir = sorted(x.name for x in destdir.iterdir())
    assert ["file%s.txt" % i for i in range(10)] == lsdir


def test_compress_uncompress_tar_modes(tmp_path):
    tocompress = tmp_path / "compressme"
    tocompress.mkdir()

    fpath = tocompress / "text.txt"
    fpath.write_text("echo foo")
    fpath.chmod(0o644)

    fpath = tocompress / "executable.sh"
    fpath.write_text("echo foo")
    fpath.chmod(0o755)

    tarfile = tmp_path / "archive.tar"
    tarball.compress(str(tarfile), "tar", str(tocompress))

    destdir = tmp_path / "destdir"
    tarball.uncompress(str(tarfile), str(destdir))

    (executable_path, text_path) = sorted(destdir.iterdir())
    assert text_path.stat().st_mode == 0o100644
    assert executable_path.stat().st_mode == 0o100755


def test_uncompress_tar_failure(tmp_path, datadir):
    """Unpack inexistent tarball should fail"""
    tarpath = os.path.join(datadir, "archives", "inexistent-archive.tar.Z")

    assert not os.path.exists(tarpath)

    with pytest.raises(FileNotFoundError, match="No such file or directory"):
        tarball.uncompress(tarpath, tmp_path)


def test_uncompress_tar(tmp_path, datadir):
    """Unpack supported tarball into an existent folder should be ok"""
    filename = "groff-1.02.tar.Z"
    tarpath = os.path.join(datadir, "archives", filename)

    assert os.path.exists(tarpath)

    extract_dir = os.path.join(tmp_path, filename)

    tarball.uncompress(tarpath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0


def test_uncompress_tarzst(tmp_path, datadir):
    """Unpack supported zst tarball into an existent folder should be ok"""
    filename = "ca-certificates-20210603-1-any.pkg.tar.zst"
    tarpath = os.path.join(datadir, "archives", filename)

    assert os.path.exists(tarpath)

    extract_dir = os.path.join(tmp_path, filename)

    tarball.uncompress(tarpath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0


def test_register_new_archive_formats(prepare_shutil_state):
    """Registering new archive formats should be fine"""
    unpack_formats_v1 = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        assert format_id[0] not in unpack_formats_v1

    # when
    tarball.register_new_archive_formats()

    # then
    unpack_formats_v2 = [f[0] for f in shutil.get_unpack_formats()]
    for format_id in tarball.ADDITIONAL_ARCHIVE_FORMATS:
        assert format_id[0] in unpack_formats_v2


def test_uncompress_archives(tmp_path, datadir):
    """High level call uncompression on supported archives"""
    archive_dir = os.path.join(datadir, "archives")
    archive_files = os.listdir(archive_dir)

    for archive_file in archive_files:
        archive_path = os.path.join(archive_dir, archive_file)
        extract_dir = os.path.join(tmp_path, archive_file)
        tarball.uncompress(archive_path, dest=extract_dir)
        assert len(os.listdir(extract_dir)) > 0


def test_normalize_permissions(tmp_path):
    for perms in range(0o1000):
        filename = str(perms)
        file_path = tmp_path / filename
        file_path.touch()
        file_path.chmod(perms)

    # add directory without any permission
    dir_path = tmp_path / "dir"
    dir_path.mkdir(mode=stat.S_IFDIR)

    for file in tmp_path.iterdir():
        if file.is_dir():
            assert file.stat().st_mode == stat.S_IFDIR
        else:
            assert file.stat().st_mode == 0o100000 | int(file.name)

    tarball.normalize_permissions(str(tmp_path))

    for file in tmp_path.iterdir():
        if file.is_dir():
            assert file.stat().st_mode == stat.S_IFDIR + 0o000755
        else:
            if int(file.name) & 0o100:  # original file was executable for its owner
                assert file.stat().st_mode == 0o100755
            else:
                assert file.stat().st_mode == 0o100644


def uncompress_archive_test(archive_path, tmp_path):
    assert os.path.exists(archive_path)

    extract_dir = os.path.join(tmp_path, "unpack", os.path.basename(archive_path))
    tarball.uncompress(archive_path, extract_dir)
    assert len(os.listdir(extract_dir)) > 0


def test_uncompress_tar_fallback_when_unpack_archive_failed(tmp_path):
    """shutil.unpack_archive can fail unpacking a tarball with missing
    directory bit set on folder contained in it while the tar command
    succeeds to perform the same task so check the fallback of using
    tar command works for such edge case.
    """
    archive_path = os.path.join(tmp_path, "repro.tar.gz")
    tf = tarfile.open(archive_path, "w:gz")
    ti = tarfile.TarInfo("dir")
    ti.mode = 0o777
    ti.type = tarfile.DIRTYPE
    tf.addfile(ti)
    ti = tarfile.TarInfo("dir/file")
    tf.addfile(ti, io.BytesIO(b"hello world"))
    tf.close()
    uncompress_archive_test(archive_path, tmp_path)


def test_uncompress_zip_imploded(tmp_path, datadir):
    """Unpack a zip archive with compression type 6 (implode),
    not supported by python zipfile module.
    """
    archive_path = os.path.join(datadir, "archives", "msk316src.zip")
    uncompress_archive_test(archive_path, tmp_path)


def test_uncompress_upper_archive_extension(tmp_path, datadir):
    """Copy test archives in a temporary directory but turn their names
    to uppercase, then check they can be successfully extracted.
    """
    archives_path = os.path.join(datadir, "archives")
    archive_files = [
        f
        for f in os.listdir(archives_path)
        if os.path.isfile(os.path.join(archives_path, f))
    ]
    for archive_file in archive_files:
        archive_file_upper = os.path.join(tmp_path, archive_file.upper())
        extract_dir = os.path.join(tmp_path, archive_file)
        shutil.copy(os.path.join(archives_path, archive_file), archive_file_upper)
        tarball.uncompress(archive_file_upper, extract_dir)
        assert len(os.listdir(extract_dir)) > 0


def test_uncompress_archive_no_extension(tmp_path, datadir):
    """Copy test archives in a temporary directory but turn their names
    to their md5 sums, then check they can be successfully extracted.
    """
    archives_path = os.path.join(datadir, "archives")
    archive_files = [
        f
        for f in os.listdir(archives_path)
        if os.path.isfile(os.path.join(archives_path, f))
    ]
    for archive_file in archive_files:
        archive_file_path = os.path.join(archives_path, archive_file)
        with open(archive_file_path, "rb") as f:
            md5sum = hashlib.md5(f.read()).hexdigest()
        archive_file_md5sum = os.path.join(tmp_path, md5sum)
        extract_dir = os.path.join(tmp_path, archive_file)
        shutil.copy(archive_file_path, archive_file_md5sum)
        tarball.uncompress(archive_file_md5sum, extract_dir)
        assert len(os.listdir(extract_dir)) > 0


def test_uncompress_tar_incorrect_extension(tmp_path, datadir):
    """Unpack tarball with an incorrect extension should be ok"""
    filename = "groff-1.02.tar.Z"
    orig_tarpath = os.path.join(datadir, "archives", filename)

    tarpath = os.path.join(tmp_path, filename.replace(".tar.Z", ".tar.gz"))
    shutil.copyfile(orig_tarpath, tarpath)
    assert os.path.exists(tarpath)

    extract_dir = os.path.join(tmp_path, filename)
    tarball.uncompress(tarpath, extract_dir)

    assert len(os.listdir(extract_dir)) > 0
